#include <stdint.h>
#include <ArtnetWiFi.h>
#include <FastLED.h>

#define LED_1 2
#define LED_2 15

#define PWM_PIN_0   27
#define PWM_PIN_1   26
#define PWM_PIN_2   25
#define PWM_PIN_3   33
#define PWM_PIN_4   32
#define PWM_PIN_5   21
#define PWM_PIN_6   22
#define PWM_PIN_7   23


//setting PWM properties
#define FREQ  1000
#define RES   12
uint16_t pwmRedValue        = 0;
uint16_t pwmGreenValue      = 0;
uint16_t pwmBlueValue       = 0;
uint16_t pwmUvValue         = 0;
uint16_t pwmColdWhiteValue  = 0;
uint16_t pwmWarmWhiteValue  = 0;

//DMX channels
#define DMX_START 1
uint8_t masterDimmer    = 0; //CHAN0: Master dimmer [0:255]
uint8_t rgbColorHue     = 0; //CHAN1: RGB Color
uint8_t rgbDimmer       = 0; //CHAN2: RGB Dimmer [0:255]
uint8_t coldWhiteDimmer = 0; //CHAN3: Cold White Dimmer [0:255]
uint8_t warmWhiteDimmer = 0; //CHAN4: Warm White Dimmer [0:255]
uint8_t uvDimmer        = 0; //CHAN5: UV Dimmer [0:255]
uint8_t strobMode       = 0; //CHAN6: Strob mode 
uint8_t strobFrequency  = 0; //CHAN7: Strob Frequency
uint8_t strobFlashTime  = 0; //CHAN8: Strob Flash Time



const char* ssid = "YourWifiSSID";
const char* pwd = "YourWifiPassword";

const IPAddress ip(192, 168, 1, 201);
const IPAddress gateway(192, 168, 1, 1);
const IPAddress subnet(255, 255, 255, 0);

ArtnetWiFiReceiver artnet;

hw_timer_t* strobTimer = NULL;
uint64_t strobLightOnTicks = 0;
uint64_t  strobLightOffTicks = 0; 
bool strobLightOn = false;

void IRAM_ATTR strobCallback() {
  strobLightOnTicks = computeStrobOnTicks();
  strobLightOffTicks = computeStrobOffTicks();

  timerAlarmWrite(strobTimer, strobLightOn?strobLightOffTicks:strobLightOnTicks, true);
  strobLightOn = !strobLightOn;
}

void setup() {  
  Serial.begin(115200);

  pinMode(LED_1, OUTPUT);
  digitalWrite(LED_1, LOW);
  
  pinMode(LED_2, OUTPUT);
  digitalWrite(LED_2, LOW);
  
  ledcAttachPin(PWM_PIN_0, 0);
  ledcAttachPin(PWM_PIN_1, 1);
  ledcAttachPin(PWM_PIN_2, 2);
  ledcAttachPin(PWM_PIN_3, 3);
  ledcAttachPin(PWM_PIN_4, 4);
  ledcAttachPin(PWM_PIN_5, 5);
  

  ledcSetup(0, FREQ, RES);
  ledcSetup(1, FREQ, RES);
  ledcSetup(2, FREQ, RES);
  ledcSetup(3, FREQ, RES);
  ledcSetup(4, FREQ, RES);
  ledcSetup(5, FREQ, RES);

  flushPwm();

  WiFi.begin(ssid, pwd);
  WiFi.config(ip, gateway, subnet);
  while (WiFi.status() != WL_CONNECTED) {
    digitalWrite(LED_2, HIGH);
    Serial.print(".");
    delay(250);
    digitalWrite(LED_2, LOW);
    delay(250);
  }
  Serial.print("WiFi connected, IP = ");
  Serial.println(WiFi.localIP());

  strobTimer = timerBegin(0, 80, true);
  timerAttachInterrupt(strobTimer, &strobCallback, true);
  timerAlarmDisable(strobTimer);
  
  artnet.begin();
  artnet.subscribe(0, artNetCallback);  

}

void loop() {
  artnet.parse();
  updateLight();
  flushPwm();
}

void artNetCallback(const uint8_t* data, const uint16_t dataSize) {  
   digitalWrite(LED_1, HIGH);
	for (size_t iChan = 1; iChan <= dataSize; ++iChan) {
		if(iChan < DMX_START) continue;
		else if(iChan == (DMX_START))   masterDimmer = data[iChan-1];
		else if(iChan == (DMX_START+1)) rgbColorHue = data[iChan-1];
		else if(iChan == (DMX_START+2)) rgbDimmer = data[iChan-1];
		else if(iChan == (DMX_START+3)) coldWhiteDimmer = data[iChan-1];
		else if(iChan == (DMX_START+4)) warmWhiteDimmer = data[iChan-1];
		else if(iChan == (DMX_START+5)) uvDimmer = data[iChan-1];
		else if(iChan == (DMX_START+6)) strobMode = data[iChan-1];
		else if(iChan == (DMX_START+7)) strobFrequency = data[iChan-1];
		else if(iChan == (DMX_START+8)) strobFlashTime = data[iChan-1];
		else if(iChan > (DMX_START+8)) break;
   }   
   updateStrob();     
   digitalWrite(LED_1, LOW);
}


void updateLight(){
  //UV Light
  pwmUvValue = mergeDimmer(masterDimmer, uvDimmer);

  //RGB
  CHSV hsv = CHSV(rgbColorHue, 255, rgbDimmer);  
  CRGB rgb;
  hsv2rgb_rainbow( hsv, rgb);
  pwmRedValue = mergeDimmer(masterDimmer, rgb.r);
  pwmGreenValue = mergeDimmer(masterDimmer, rgb.g);
  pwmBlueValue = mergeDimmer(masterDimmer, rgb.b);

  //Whites
  pwmColdWhiteValue = (strobMode > 0 && !strobLightOn)?0:mergeDimmer(masterDimmer, coldWhiteDimmer);
  pwmWarmWhiteValue = mergeDimmer(masterDimmer, warmWhiteDimmer);


}

void updateStrob(){
  bool isTimerAlarmEnabled = timerAlarmEnabled(strobTimer);
  
  if(strobMode == 0 && isTimerAlarmEnabled){
    timerAlarmDisable(strobTimer);
  }
  else if (strobMode > 0 && !isTimerAlarmEnabled) {
    strobLightOnTicks = computeStrobOnTicks();
    strobLightOn = true;
    timerAlarmWrite(strobTimer, strobLightOnTicks, true);
    timerAlarmEnable(strobTimer);
  }
}

void flushPwm(){
  ledcWrite(0, pwmColdWhiteValue);
  ledcWrite(1, pwmUvValue);
  ledcWrite(2, pwmRedValue);
  ledcWrite(3, pwmGreenValue);
  ledcWrite(4, pwmBlueValue);
  ledcWrite(5, pwmWarmWhiteValue);
}


uint16_t mergeDimmer(uint8_t master, uint8_t local){
  uint16_t merged = master * local;
  merged = merged >> (16-RES);
  return merged;
}

uint64_t computeStrobOnTicks(){
  uint64_t onTicks = 1000000 / (strobFrequency/8+1); //Full period
  onTicks = onTicks * (25+strobFlashTime/5) / 100;  
  return onTicks;
}

uint64_t computeStrobOffTicks(){
  uint64_t onTicks = 1000000 / (strobFrequency/8+1); //Full period
  onTicks = onTicks * (75-strobFlashTime/5) / 100;  
  return onTicks;
}

void printDebugDmx(){
  Serial.println("DMX:");
  Serial.print("   1: masterDimmer: ");
  Serial.println(masterDimmer);

  Serial.print("   2: rgbColorHue: ");
  Serial.println(rgbColorHue);

  Serial.print("   3: rgbDimmer: ");
  Serial.println(rgbDimmer);

  Serial.print("   4: coldWhiteDimmer: ");
  Serial.println(coldWhiteDimmer);

  Serial.print("   5: warmWhiteDimmer: ");
  Serial.println(warmWhiteDimmer);

  Serial.print("   6: uvDimmer: ");
  Serial.println(uvDimmer);

  Serial.print("   7: strobMode: ");
  Serial.println(strobMode);

  Serial.print("   8: strobFrequency: ");
  Serial.println(strobFrequency);

  Serial.print("   9: strobFlashTime: ");
  Serial.println(strobFlashTime);

  Serial.println("---------------------");
}

void printDebugPwm(){
  Serial.println("PWM:");
  Serial.print("   1: pwmRedValue: ");
  Serial.println(pwmRedValue);

  Serial.print("   2: pwmGreenValue: ");
  Serial.println(pwmGreenValue);

  Serial.print("   3: pwmBlueValue: ");
  Serial.println(pwmBlueValue);

  Serial.print("   4: pwmUvValue: ");
  Serial.println(pwmUvValue);

  Serial.print("   5: pwmColdWhiteValue: ");
  Serial.println(pwmColdWhiteValue);

  Serial.print("   6: pwmWarmWhiteValue: ");
  Serial.println(pwmWarmWhiteValue);

  Serial.println("---------------------");
}
